#!/usr/bin/python
# -*- coding: utf-8 -*-

##   Copyright 2010 Pierre-Nicolas <pnprog@tuxfamily.org>

##   File: goban.py
##   This file is part of GoSP 0.01.03, a program for training
##   at go with GNU Go.
##   Author: Pierre-Nicolas <pnprog@tuxfamily.org>
##   Date: Sun, 21 Mar 2010 20:28:14 +0800
##   For more information, see http://gosp.tuxfamily.org/

##   GoSP is free software: you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   GoSP is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.
##
##   You should have received a copy of the GNU General Public License
##   along with GoSP.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding:Utf-8 -*-
from sys import exit
from gtp import gtp
from copy import deepcopy
from random import shuffle, random

try:import Tkinter as Tk
except:
	print "Can not import \"Tkinter\" module. Is \"python-tk\" installed?"
	sys.exit()

class goban():
	def __init__(self,canvas,size=19):
		self.size=size
		self.canvas=canvas
		self.grid=[]
		for i in range(size):
			self.grid.append([0]*size)
		
		self.history=[]
		self.history.append(deepcopy(self.grid))
		
		self.last_white=[-1,-1]
		self.last_black=[-1,-1]
		
		self.stones=[]
		self.space=0
		self.border=0
		
		
		try:self.gnugo=gtp(("gnugo", "--mode", "gtp"))
		except:
			print "Can not run GnuGo. Is GnuGo installed?"
			exit()
		self.gnugo.boardsize(size)
		self.gnugo.reset()
		
		#those 2 variables will change to Tk.IntVar later
		self.auto_answer=1
		self.one_color_go=1
		
		self.mistake_rate=0.05
		
		
	def neighborhood(self,i,j):
		list=[]
		if 0 <= i+1 <= self.size-1 and 0 <= j <= self.size-1:
			list.append([i+1,j])

		if 0 <= i-1 <= self.size-1 and 0 <= j <= self.size-1:
			list.append([i-1,j])
		
		if 0 <= i <= self.size-1 and 0 <= j-1 <= self.size-1:
			list.append([i,j-1])
		
		if 0 <= i <= self.size-1 and 0 <= j+1 <= self.size-1:
			list.append([i,j+1])
		return list
	
	def free_neighborhood(self,i,j):
		list=[]
		for x,y in self.neighborhood(i,j):
			if self.grid[x][y]==0:
				list.append([x,y])
		return list
	
	def remove_group(self,i,j):
		#print i,j
		color=self.grid[i][j]
		self.grid[i][j]=0
		
		for x,y in self.neighborhood(i,j):
			if self.grid[x][y]==color:
				self.remove_group(x,y)


	def countlib(self,i,j,lib=0,tab=None):
		if not tab:
			start=True
			tab=[]
			for p in range(self.size):tab.append([0]*self.size)
			#print tab
		else:start=False
		color=self.grid[i][j]
		#print "c=",color
		if color==0:
			return -1
		tab[i][j]=1
		
		for x,y in self.neighborhood(i,j):
			if self.grid[x][y]==color and tab[x][y]==0:
				lib,tab=self.countlib(x,y,lib,tab)
			elif self.grid[x][y]==0 and tab[x][y]==0:
				tab[x][y]=2
				lib+=1
		
		if start:
			return lib
		else:
			return lib,tab
	
	def mistake(self,i,j):
		options=[[i,j]]+self.free_neighborhood(i,j)
		shuffle(options)
		return options[0]
		
	
	def gnugo_play_black(self,event=None):
		move=self.gnugo.play_black()
		if move not in ["pass","resign","PASS"]:
			i,j=self.move2ij(move)
			if random()<self.mistake_rate:
				ii,jj=self.mistake(i,j)
				self.gnugo.undo()
				if self.gnugo.place_black(self.ij2move(ii,jj)):
					i,j=ii,jj
					print ";-)"
				else:
					move=self.gnugo.play_black()
					i,j=self.move2ij(move)
					
			self.place_black(i,j)
			self.grid[i][j]=1
			self.history.append(deepcopy(self.grid))
			self.last_black=[i,j]
			self.display()
		else:self.alert(move)
	
	def gnugo_play_white(self,event=None):
		move=self.gnugo.play_white()
		if move not in ["pass","resign","PASS"]:
			i,j=self.move2ij(move)
			if random()<self.mistake_rate:
				ii,jj=self.mistake(i,j)
				self.gnugo.undo()
				if self.gnugo.place_white(self.ij2move(ii,jj)):
					i,j=ii,jj
					print ";-)"
				else:
					move=self.gnugo.play_white()
					i,j=self.move2ij(move)
			self.place_white(i,j)
			self.last_white=[i,j]
			self.grid[i][j]=2
			self.history.append(deepcopy(self.grid))
			self.display()
		else:self.alert(move)
	
	def place_black(self,i,j):
		color=2 #color of the opposite side
		
		for x,y in self.neighborhood(i,j):
			if self.grid[x][y]==color:
				if self.countlib(x,y)==1:
					#will be killed
					self.remove_group(x,y)

	def place_white(self,i,j):
		color=1 #color of the opposite side
		
		for x,y in self.neighborhood(i,j):
			if self.grid[x][y]==color:
				if self.countlib(x,y)==1:
					#will be killed
					self.remove_group(x,y)

	def button1(self,event):
		#add/remove black stone
		#check pointer location
		i,j=self.xy2ij(event.x,event.y)
		if 0 <= i <= self.size-1 and 0 <= j <= self.size-1:
			#inside the grid
			#what is under the pointer ?
			
			if self.grid[i][j]==0:
				#nothing, so we add a black stone
				if self.gnugo.place_black(self.ij2move(i,j)):
					self.place_black(i,j)
					self.last_black=[i,j]
					self.grid[i][j]=1
					self.history.append(deepcopy(self.grid))
				
					self.display()
				
					if self.auto_answer.get()==1:
						self.gnugo_play_white()
				return

	def button2(self,event):
		#undo
		self.undo()

	def button3(self,event):
		#add/remove white stone
		#check pointer location
		i,j=self.xy2ij(event.x,event.y)
		if 0 <= i <= self.size-1 and 0 <= j <= self.size-1:
			#inside the grid
			#what is under the pointer ?
			
			if self.grid[i][j]==0:
				#nothing, so we add a white stone
				if self.gnugo.place_white(self.ij2move(i,j)):
					self.place_white(i,j)
					self.last_white=[i,j]
					self.grid[i][j]=2
					self.history.append(deepcopy(self.grid))
					self.display()
				
					if self.auto_answer.get()==1:
						self.gnugo_play_black()
				return
	
	def undo(self):
		try:
			#print self.gnugo.show_board()
			if self.gnugo.undo():
				self.history=self.history[:-1]
				self.grid=deepcopy(self.history[-1])
				self.last_white=[-1,-1]
				self.last_black=[-1,-1]
				self.display()
				#print self.gnugo.show_board()
			else: print "GNUGO could not undo"
		except:
			print "Undo failed"
			pass
	
	def draw_point(self,i,j,radius,color="black"):
		x,y=self.ij2xy(i,j)
		self.canvas.create_oval(x-radius,y-radius,x+radius,y+radius,fill=color)
		
	def draw_stone(self,i,j,color):
		self.draw_point(i,j,self.space/2.5,color)
		
		
	
	def ij2xy(self,x,y):
		y=self.size-y-1
		return (x+self.border)*self.space,(y+self.border)*self.space
	
	def xy2ij(self,x,y):
		return int(round(x/self.space))-self.border,self.size-(int(round(y/self.space))-self.border)-1

	def ij2move(self,i,j):
		letters=['A','B','C','D','E','F','G','H','J','K','L','M','N','O','P','Q','R','S','T']
		return letters[i]+str(j+1)
	
	def move2ij(self,move):
		#print "move2ij",move
		letters=['A','B','C','D','E','F','G','H','J','K','L','M','N','O','P','Q','R','S','T']
		return letters.index(move[0]),int(move[1:])-1
	
	def display(self,event=None):
		if event:
			#nouvel affichage provoqué par un redimensionnement
			width=event.width
			height=event.height
			self.canvas_width=width
			self.canvas_height=height
		else:
			width=self.canvas_width
			height=self.canvas_height
		
		#suppression de l'ancien canvas	
		for item in self.canvas.find_all():
			self.canvas.delete(item)
		
		#line drawing and numbering
		letters=['A','B','C','D','E','F','G','H','J','K','L','M','N','O','P','Q','R','S','T']
		b=1 #border size in grid square size
		self.border=b
		
		space=min(width,height)/(self.size+2*b-1.) #space betwen two lines
		self.space=space
		i=0
		for line in self.grid:
			self.canvas.create_text((i+b)*space,(b-0.5)*space, text=letters[i],font=("Helvectica", str(int(space/3))))
			self.canvas.create_line((i+b)*space,b*space,(i+b)*space,(self.size+b-1)*space,fill="black")
			self.canvas.create_text((i+b)*space,(self.size+b-0.5)*space, text=letters[i],font=("Helvectica", str(int(space/3))))
			
			self.canvas.create_text((b-0.5)*space,(i+b)*space, text=str(self.size-i),font=("Helvectica", str(int(space/3))))
			self.canvas.create_line(b*space,(i+b)*space,(self.size+b-1)*space,(i+b)*space,fill="black")
			self.canvas.create_text((self.size+b-0.5)*space,(i+b)*space, text=str(self.size-i),font=("Helvectica", str(int(space/3))))
			i+=1

		#hoshis
		r=self.space/7
		if self.size==19:
			for i,j in [[3,3],[3,9],[9,9],[3,15],[15,15],[15,9],[9,15],[15,3],[9,3]]:
				self.draw_point(i,j,r,"black")
				#x,y=self.ij2xy(i,j)
				#self.canvas.create_oval(x-e,y-e,x+e,y+e,fill="black")
		
		#stones drawing
		
		if self.one_color_go.get()==0:b_color,w_color="black","white"
		else:b_color,w_color="grey","grey"
		
		for i in range(self.size):
			for j in range(self.size):
				if self.grid[i][j]==1:
					#black
					self.draw_stone(i,j,b_color)
				if self.grid[i][j]==2:
					self.draw_stone(i,j,w_color)
		
		#last play
		r=self.space/10
		[i,j]=self.last_black
		if i>=0:self.draw_point(i,j,r,"blue")
		[i,j]=self.last_white
		if i>=0:self.draw_point(i,j,r,"red")	
		#self.canvas.create_oval(x-e,y-e,x+e,y+e,fill="red")	
		
		self.canvas.update_idletasks()
	
	def end_game(self):
		self.display()
		white_territory=0
		black_territory=0
		dead_white=0
		dead_black=0
		
		r=self.space/7
		for i in range(self.size):
			for j in range(self.size):
				status=self.gnugo.final_status(self.ij2move(i,j))
				if status=="white_territory":
					white_territory+=1
					self.draw_point(i,j,r,"white")
					
				if status=="black_territory":
					black_territory+=1
					self.draw_point(i,j,r,"black")
				
				if status=="dead" and self.grid[i][j]==1:
					dead_black+=1
					white_territory+=1
					self.draw_point(i,j,r,"white")
						
				if status=="dead" and self.grid[i][j]==2:
					dead_white+=1
					black_territory+=1
					self.draw_point(i,j,r,"black")

		result=self.gnugo.final_score()+" :\n"
		result+="Black= "+str(black_territory)+" territory + "+str(dead_white)+" dead white stones\n"
		result+="White= "+str(white_territory)+" territory + "+str(dead_black)+" dead black stones + komi\n"
		self.alert(result)
	
	
	def alert(self,text_to_display):
		popup=Tk.Toplevel()
		label= Tk.Label(popup,text=text_to_display)
		label.pack()
		
		ok_button = Tk.Button(popup, text="OK", command=popup.destroy)
		ok_button.pack()
		
		popup.mainloop()
	
